var d3 = require('d3');
require('d3.chart');
require('@domoinc/notifier');
var _ = require('lodash');
//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
// Base Widget:
// This widget doesn't do anything by it self but it contains all the "base"
// features that our widgets need. This widget should be "extended" to do anthing that
// you need it to do.
//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
module.exports = d3.chart('BaseWidget', {


  //**********************************************************************************
  // Anything in the initialization method will only run once.
  // i.e. It will not be run every time you call draw.
  //**********************************************************************************
  initialize: function () {
    "use strict";
    // Make the chart constructor accessible to all other methods.
    var _Chart = this;




    //----------------------------------------------------------------------------------
    // Data validation function and chart variable setup.
    //----------------------------------------------------------------------------------
    _Chart.transform = function (data) {
      _Chart.updateSizeableConfigs();
      var validData = _Chart.validateData(data);

      return validData;
    };




    //----------------------------------------------------------------------------------
    // Default Values for the wiget's configurable options.
    // Valid types for UI elements are: string, number, color, select, boolean
    //----------------------------------------------------------------------------------
    _Chart._config = {
      height: {
        name: 'Chart Height',
        description: '',
        category: 'Dimensions',
        type: 'number',
        value: 250,
        units: 'px'
      },
      width: {
        name: 'Chart Width',
        description: '',
        category: 'Dimensions',
        type: 'number',
        value: 250,
        units: 'px'
      },
      shouldValidate: {
        description: 'Flag for turning off data validation',
        type: 'boolean',
        value: true
      },
      updateSizeableConfigs: {
        description: 'Flag for turning off the mimic of illustrator\'s scale functionality',
        type: 'boolean',
        value: true
      }
    };




    //----------------------------------------------------------------------------------
    // Data Definition:
    // Set functions on how to access and validate data values.
    //----------------------------------------------------------------------------------
    // This should be set in the "extended chart".
    _Chart._dataDefinition = {};




    //----------------------------------------------------------------------------------
    // Static - Anything that is defined here will never change
    //----------------------------------------------------------------------------------
    _Chart._layerGroup = this.base.append('g');




    //----------------------------------------------------------------------------------
    // Notifier:
    // This must come after any other layers/groups
    //----------------------------------------------------------------------------------
    _Chart._notifier = this.base.append('g')
      .attr('class', 'notifier')
      .chart('Notifier');

    _Chart._notifier
      .width(function () {
        return _Chart.config('width');
      });

    _Chart._notifier
      .height(function () {
        return _Chart.config('height');
      });
  },




  //----------------------------------------------------------------------------------
  // Helper Functions
  //----------------------------------------------------------------------------------
  /**
   * Merges the extending chart's config with the base's config
   * @param  {object} newConfig The extending chart's config
   */
  mergeConfig: function mergeConfig(newConfig) {
    "use strict";
    var _Chart = this;
    if (_Chart._config) {
      _Chart._config = _.extend(_Chart._config, newConfig);
    }
    else {
      _Chart._config = newConfig;
    }
  },


  /**
   * Merges the extending chart's data definitions with the base's data definition
   * @param  {object} newDefinition The extending chart's data definition
   */
  mergeDataDefinition: function mergeDataDefinition(newDefinition) {
    "use strict";
    var _Chart = this;
    if (_Chart._dataDefinition) {
      _Chart._dataDefinition = _.extend(_Chart._dataDefinition, newDefinition);
    }
    else {
      _Chart._dataDefinition = newDefinition;
    }
  },


  /**
   * Runs data though your validation functions and throws out bad data.
   * @param  {Array} data The data you passed to the draw method of your chart.
   * @return {Array}      The validated data.
   */
  validateData: function (data) {
    var validData, _Chart = this;
    _Chart._notifier.clearDataValidationMessages();
    if (_Chart.c('shouldValidate')) {
      if (!data || !_.flatten(data).length) {
        _Chart._notifier.appendMessage(_Chart.config('chartName'), 'NO_DATA', 'No Data', _Chart._dataDefinition, data);
        validData = [];
      }
      else {
        validData = data.filter(function (line) {
          for (var col in _Chart._dataDefinition) {
            if (!_Chart._dataDefinition[col].validate(line)) {
              _Chart._notifier.appendMessage(_Chart.c('chartName'), 'WARN', 'Invalid value at column for ' + col, _Chart._dataDefinition[col], line);
              return false;
            }
          }
          return true;
        });
        if (!validData.length) {
          _Chart._notifier.appendMessage(_Chart.config('chartName'), 'NO_DATA', 'No Data', _Chart._dataDefinition, data);
        }
      }
    }
    else {
      validData = data;
    }
    _Chart._notifier.draw();
    return validData;
  },




  //----------------------------------------------------------------------------------
  // Public Functions:
  // Getter and Setter functions are documented here.
  //----------------------------------------------------------------------------------


  /**
   * Sets one or many of the data definition properties for the chart
   * @param  {object} obj Contains data definiton objects you would like to set or update.
   * @return {object}     Returns the chart for chainability.
   */
  dataDefinition: function (obj) {
    "use strict";

    if (arguments.length === 0) {
      return this._dataDefinition;
    }

    for (var key in obj) {
      if (this._dataDefinition[key]) {
        for (var subKey in obj[key]) {
          if (obj[key].hasOwnProperty(subKey)) {
            this._dataDefinition[key][subKey] = obj[key][subKey];
          }
        }
      }
      else {
        console.warn(key + ' is not a valid data property.');
      }
    }

    return this;
  },


  /**
   * This will get or set any of the chart's configuration options defined
   * in the 'chart._config' object.
   * @param  {String or Object}   item        If string, it will return the value for that config item.
   *
   * If object, it will update that config item with set value.
   * @param  {String or Number}   [value]     The value to update config item with.
   * @return {object}                         The chart to preserve chainability.
   */
  config: function (item, value) {
    "use strict";

    if (item === undefined) {
      return this;
    }
    if (value === undefined && typeof item === 'string') {
      return this._config[item] ? this._config[item].value : undefined;
    }

    var items;
    if (typeof item === 'object') {
      items = item;
    }
    else {
      items = {};
      items[item] = value;
    }

    for (var key in items) {
      if (!this._config[key]) {
        console.warn(this.c('chartName') + ':' + key + ' is not a valid config property.');
      }
      else {
        this._config[key].value = items[key];

        if (this._config[key].updatePercent) {
          this._config[key].updatePercent();
        }

        if (this._config[key].onChange) {
          console.warn(this.c('chartName') + ':' + key + ': Config\'s onChange function is deprecated and will be removed shortly.');
          this._config[key].onChange.call(this, items[key]);
        }
      }
    }

    return this;
  },


  /**
   * This will get or set any of the chart's accessors defined
   * in the 'chart._dataDefinition' object.
   * @param  {String or Object}   item        If string, it will return the function for that accessor item. If object, it will update that accessor with set function.
   * @param  {function}           [value]     The function to update accessor item with.
   * @return {object}                         The chart to preserve chainability.
   */
  accessors: function (item, value) {
    "use strict";

    if (item === undefined) {
      return this;
    }
    if (value === undefined && typeof item === 'string') {
      if (this._dataDefinition[item]) {
        return this._dataDefinition[item].accessor;
      }
      else {
        console.warn(item + ' is not a valid accessor.');
        return d3.scale.identity();
      }
    }

    var items;
    if (typeof item === 'object') {
      items = item;
    }
    else {
      items = {};
      items[item] = value;
    }

    for (var key in items) {
      if (this._dataDefinition[key]) {
        this._dataDefinition[key].accessor = items[key];
      }
      else {
        console.warn(key + ' is not a valid accessor.');
      }
    }

    return this;
  },


  /**
   * Shorthand for config getter/setter
   */
  c: function (item, value) {
    "use strict";
    return this.config.apply(this, arguments);
  },


  /**
   * Shorthand for accessor getter function
   */
  a: function (item, value) {
    "use strict";
    return this.accessors.apply(this, arguments);
  },


  /**
   * Visits each config element.
   * If a config element has a size function it will update that configs value
   * with the returned result of its size function.
   */
  updateSizeableConfigs: function () {
    if (!this.c('updateSizeableConfigs')) return;

    var configKeys = Object.keys(this._config);

    var configElement = null;
    for (var i = 0; i < configKeys.length; i++) {
      configElement = this._config[configKeys[i]];

      if (configElement.size !== undefined) {
        configElement.value = configElement.size();
      }
    }
  },


  /**
   * Takes a theme obj and applies it to the widget's config elements.
   *
   * @param themeObj {Object} {"name": "Dark Bold", "domoDefinedTheme": true, "config": { gaugeFillPrimaryColor: '#1C5CAE'} };
   */
  applyTheme: function (themeObj) {
    if (themeObj && themeObj.config) {

      var applyThemeConfigs = themeObj.config;
      var widgetConfigs     = this._config;

      for (var widgetConfigElementName in widgetConfigs) {

        var hasValidThemeElement = (widgetConfigs.hasOwnProperty(widgetConfigElementName) &&
                                    widgetConfigs[widgetConfigElementName].theme !== undefined);

        if (hasValidThemeElement) {

          if (applyThemeConfigs[widgetConfigs[widgetConfigElementName].theme]) {
            widgetConfigs[widgetConfigElementName].value = applyThemeConfigs[widgetConfigs[widgetConfigElementName].theme];
          }

        }

      }

      return true;
    }

    return false;
  }
});

